##########################################################
#
#  fa_re: 
#
#     Finite Automata & Regular Expressions package.
#
#                               (C) Pavel Sutyrin, 2005. 
#
##########################################################

fa_re := module ()
  export 
         fa_find_sc,   # finds connected components
         fa_find_ssc,  # finds strongly-connected components

         fa_prn_m,     # these two utilities print M- and N- numerations of vertices,
         fa_prn_n      # that were assigned during last depth search
         ;
  local
         reset_nums,   # resets M- and N-numeration.
         fa_ds_single, # single depth search, builds single span tree
         fa_ds_full    # full depth search, builds complete span forest
         ;

  option package; 

fa_find_sc := proc( G )
  return fa_ds_full( G );
end proc:


fa_find_ssc := proc( G )
  local n_list, n, ssc_set;

  ssc_set := [];

  ## first search, forward
  fa_ds_full( G, forward );

  # sort vertices by N-numbers descending 
  n_list := sort( op( 2, G( N_Num )), 
              (x,y) -> rhs( x ) > rhs( y ) ); 


  ## second search, reverse
  reset_nums( G );
  for n in n_list 
  do
    if( not assigned( G( M_Num )[ lhs( n ) ] ) )
    then 
      ssc_set := [ op( ssc_set ), fa_ds_single( G, lhs( n ), reverse ) ];
    end if;
  end do;
  
  return ssc_set;

end proc:

reset_nums := proc( G )
  G ( M_Num ) := table();
  G ( N_Num ) := table();
end proc:

fa_ds_full := proc( G, rev_flag )
  local vlist, i, cc_set;
  vlist := [ op( networks[ vertices ]( G ) ) ];

  cc_set := [];
  reset_nums( G );

  for i from 1 to nops( vlist ) do
    if( not assigned( G( M_Num )[ vlist[ i ] ] ) )
    then

      # find a spanning tree from this vertex
      if( nargs >=2 ) 
      then 
        cc_set := [ op( cc_set ), fa_ds_single( G, vlist[ i ], rev_flag ) ];
      else 
        cc_set := [ op( cc_set ), fa_ds_single( G, vlist[ i ] ) ];
      end if;

    end if;
  end do;

  return cc_set;

end proc:

#############################################
#
# fa_ds_single:
#
# single depth search procedure: iterative, stack based version.  
# performs depth search in single connected component of G.
#
# INPUT: G - graph (of type networks/GRAPH);
#        u_0 - starting vertex.
#
#        [dir] = forward | reverse; 
#                specifies edge direction to follow,
#                if omitted, then edges are considered undirected
#
# OUTPUT: the list of vertices, including u_0, that aggregate single connected
# component. sorted in M-ascending order.
#
# SIDE EFFECTS: M-numbers are stored in N_Num table of graph;
#               N-numbers are stored in N_Num table of graph.
#
########

fa_ds_single := proc( G, u_0, dir )
  local S, m, cur_m, n, cur_n,
        u, v, vs, ws,   # u -> v -> w
        status,         # vertex status
        cc,             # connected component (vertices)
        advance         # function, either departures or arrivals or neighbors
        ; 

  description `finds M- and N- numerations of G using depth search, creates   M_Num and N_Num tables of G`;

  if( nargs < 2 )
  then
    error `please, specify starting vertex`;
  end if;

  # undirected edges by default
  advance := networks[ neighbors ];

  if( nargs >= 3 )
  then
    if( dir = reverse )
    then
      advance := networks[ arrivals ];
    elif ( dir = forward )
    then
      advance := networks[ departures ];
    end if;
  end if;

  # use existing tables or create new

  if( assigned( G( M_Num ) ) )
  then m := G( M_Num );
  else m := table();
  end if;

  if( assigned( G( N_Num ) ) )
  then n := G( N_Num );
  else n := table();
  end if;

  # continue counting M and N numbers
  cur_m := nops( op( eval( m ) ) ) + 1;
  cur_n := nops( op( eval( n ) ) ) + 1;

  # initialize stack 
  
  m[ u_0 ] := cur_m;
  cur_m := cur_m + 1;
  cc := [ u_0 ];

  # queue of u daughters to process
  vs := queue[ new ]( op( advance( u_0, G ) ) );
  S := stack[ new ]( [u_0, eval( vs )] );

  # S always contains a current search path, and each node is paired 
  # with queue of his unprocessed daughters. Thus, it is similar to 
  # runtime stack that would appear if everything is programmed recursively.

  while( not stack[ empty ]( S ) ) 
  do

    u, vs := op( stack[ pop ]( S ) );

    if( queue[ empty ]( vs ) ) # u is fully processed, assign N-number.
    then
      n[ u ] := cur_n;
      cur_n := cur_n + 1;
    else # if u has daughters
      v := queue[ dequeue ]( vs ); 

      # find first M-unvisited daughter of u
      while( assigned( m[ v ] ) and not queue[ empty ]( vs ) ) 
      do
        v := queue[ dequeue ]( vs ); 
      end do;

      if( not assigned( m[ v ] ) ) # if found
      then
        m[ v ] := cur_m; # mark u 
        cur_m := cur_m + 1;
        cc := [ op(cc), v ];
        ws := queue[ new ]( op( advance( v, G ) ) );

        # don't forget to return to u later
        stack[ push ]( [u, eval( vs )], S );
        stack[ push ]( [v, eval( ws )], S );

      else # no unvisited daughters were found
        n[ u ] := cur_n;
        cur_n := cur_n + 1;
      end if;
    end if;
  end do;

  # store M-numeration and N-numeration
  G ( M_Num ) := eval( m ); 
  G ( N_Num ) := eval( n ); 

  return cc;

end proc:

###############################################################

fa_prn_m := proc( G )
  local c, t;
  description `prints M-numbers of G, ascending`;

  if( not assigned( G (M_Num )))
  then
    error `graph has no M-numeration`;
  end if;
  t := sort( op( 2, G( M_Num )), 
              (x,y) -> rhs( x ) < rhs( y ) ); #sort by M-numbers 
  for c in t 
  do
    printf( "M( %a ) = %d\n", lhs( c ), rhs( c ) );
  end do;  

  return NULL;

end proc:

###############################################################

fa_prn_n := proc( G )
  local c, t;
  description `prints N-numbers of G, ascending`;

  if( not assigned( G (N_Num )))
  then
    error `graph has no N-numeration`;
  end if;
  t := sort( op( 2, G( N_Num )), 
              (x,y) -> rhs( x ) < rhs( y ) ); #sort by N-numbers 
  for c in t 
  do
    printf( "N( %a ) = %d\n", lhs( c ), rhs( c ) );
  end do;  

  return NULL;

end proc:

end module:

###############################################################

printf( " \nThis is Finite Automata \& Regular Expressions (fa_re) package.  \n\n" );
printf( "New functions bound:\n" );

with( fa_re );

with( networks ):
printf( "Also loaded networks package.\n" );


